/***************************************************************************
 *   Copyright (C) 2004 by Petric Frank                                    *
 *   pfrank@gmx.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "cvs-command-tree.h"
#include "cvs-session.h"
#include "cvs-connection.h"



cvs_command_tree::cvs_command_tree (cvs_directory_entry &dir, const std::string &lab)
: cvs_label (lab), root_entry (dir), current_dir (dir), started_scan (0)
{
}



cvs_command_tree::~cvs_command_tree ()
{
}



bool cvs_command_tree::run (cvs_session &session)
{
  const cvs_connection &conn = session.get_connection ();

  started_scan = time (0);
  root_entry.set_scantime (started_scan);

  // reset the log filename
  root_entry.set_logfilename ("");

  cvs_root = conn.cvsroot ();
  cvs_project = conn.project ();

  // store responses from CVS server
  if (establish_logging (session))
    root_entry.set_logfilename (filename);

  if (!send_command (session))
    return false;

  current_path = "";

  bool result = process_response (session);

  if (result)
    root_entry.cleanup_tree (started_scan);

  scan_duration = difftime (time (0), started_scan);

  return result;
}



bool cvs_command_tree::send_command (cvs_session &session)
{
  session.send () << "Argument -s" << std::endl
		  << "Argument -r" << std::endl
		  << "Argument 0" << std::endl
		  << "Argument -r" << std::endl
		  << "Argument " << cvs_label << std::endl
		  << "Argument " << cvs_project << std::endl
		  << "rdiff" << std::endl;

  return true;
}



bool cvs_command_tree::process_line (cvs_session & session, const std::string & line)
{
  const char FileMarker[]	= "M File ";
  const char DirMarker[]	= "E cvs rdiff: Diffing ";
  const char VersionMarker[]	= " is new; ";
  //                                        ^^^^^^^-- =Revision - if given in rdiff
  // Example: HEAD revision ----> " is new: HEAD revision ";
  const char RevisionMarker[]	= " revision ";
  const int FileMarkerSize	= sizeof (FileMarker) - 1;
  const int DirMarkerSize	= sizeof (DirMarker) - 1;
  const int VersionMarkerSize	= sizeof (VersionMarker) - 1;
  const int RevisionMarkerSize	= sizeof (RevisionMarker) - 1;

  // is it a file ?
  if (std::string::traits_type::compare (line.data (), FileMarker, FileMarkerSize) == 0)
  {
    std::string name = line.substr (FileMarkerSize);
    std::string version;

    std::string::size_type pos = name.find (VersionMarker + cvs_label + RevisionMarker);
    if (pos == std::string::npos)
      version = "<unknown>";
    else
    {
      version = name.substr (pos + VersionMarkerSize + cvs_label.length () + RevisionMarkerSize);
      name.erase (pos);
    }

    // now we have the data - insert it into the tree
    if (current_path.length () != 0)
      name.erase (0, current_path.length () + 1);	// kill path part

    add_file (name, version);

    return true;
  }

  // is it a directory ?
  if (std::string::traits_type::compare (line.data (), DirMarker, DirMarkerSize) == 0)
  {
    current_path = line.substr (DirMarkerSize);
    if (current_path == ".")
      current_path = "";

    // now we have the data - insert it into the tree
    add_directory (current_path, current_dir);

    return true;
  }

  // line ignored
  return true;
}



void cvs_command_tree::add_directory (const std::string &path, cvs_directory_entry & entry)
{
  cvs_directory_entry current = root_entry;
  std::string remain = path;

  while (remain.length () != 0)
  {
    std::string name = remain;
    std::string::size_type pos = name.find ('/');
    if (pos != std::string::npos)
    {
      name.erase (pos);
      remain.erase (0, pos + 1);
    }
    else
      remain.clear ();

    cvs_directory_entry item (name, "", true);
    if (current.get_entry (name, item))
    {
      // path element alrady exists - do further checks
      if (!item.is_dir ())
      {
        // oha - now we have a directory, but we had already a file with this name
	// Result: kill the file entry and allocate the directory
	current.remove_entry (name);

        item = cvs_directory_entry (name, "", true);

        current.add_entry (item);
      }
    }
    else
    {
      // is a new path element - add it
      current.add_entry (item);
    }

    current = item;
    
    item.set_scantime (started_scan);
  }
  
  entry = current;
}



void cvs_command_tree::add_file (const std::string &name, const std::string &version)
{
  cvs_directory_entry item (name, version, false);

  if (current_dir.get_entry (name, item))
  {
    // we have already an entry with this name
    if (item.is_dir ())
    {
      // bad - we have a file, but a directory is stored at this place in the tree
      // so we discard the directory
      current_dir.remove_entry (name);
      
      item = cvs_directory_entry (name, version, false);
      
      current_dir.add_entry (item);
    }
    else
    {
      if (item.version () != version)
      {
        // version hs changed - invalidate the data
	item.set_version (version);
	
	item.invalidate ();
      }
    }
  }
  else
  {
    // we have a new entry
    current_dir.add_entry (item);
  }
  
  item.set_scantime (started_scan);
}
