/***************************************************************************
 *   Copyright (C) 2004 by Petric Frank                                    *
 *   pfrank@gmx.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "cvsfs-fuse.h"
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <pwd.h>
#include <sstream>
#include "fs-layer-private.h"
#include "fs-layer-checkout.h"
#include "fs-layer-cvs.h"
#include "cvs-connection-factory.h"
#include "cvs-connection.h"
#include "cvs-repository.h"
#include "system-log.h"



cvsfs_fuse::cvsfs_fuse ()
: uid (0), gid (0), private_files (0), mount_view (), users_home ("/tmp")
{
  // set cache base
  struct passwd *pw = ::getpwuid (::geteuid ());
  if (pw)
    users_home = pw->pw_dir;
}



cvsfs_fuse::~cvsfs_fuse ()
{
}



void cvsfs_fuse::shutdown ()
{
  mount_view.shutdown ();
}



bool cvsfs_fuse::run_prepare (int &argv, char *argc[])
{
  system_log *log = system_log::instance ();
  int cache_loc = -1;
  int loop;

  for (loop = 1; loop < argv; ++loop)
  {
    if (strcmp (argc[loop], "-d") == 0)
    {
      log->info << "Debug mode enabled by option '-d'." << std::endl;

      log->debug.enable ();

      break;
    }

    if ((strcmp (argc[loop], "-o") == 0) && strstr (argc[loop + 1], "debug"))
    {
      log->info << "Debug mode enabled by option '-o debug'." << std::endl;

      log->debug.enable ();

      break;
    }

    if (strcmp (argc[loop], "-d") == 0)
    {
      cache_loc = loop;
    }
  }
  
  // set cache base
  if ((cache_loc != -1) && (cache_loc < (argv - 1)))
  {
    cache_base = argc[cache_loc + 1];

    // remove the argument
    for (loop = cache_loc + 1; loop < argv; ++loop)
      argc[loop - 2] = argc[loop];

    argv -= 2;
  }
  else
    cache_base = users_home + "/.cvsfs";

  // add layer for private files
  private_files = new fs_layer_private (cache_base + "/private");

  mount_view.set_private_layer (private_files);

  return true;
}



void cvsfs_fuse::help_options ()
{
  std::cerr << "    -c <path>           caching directory" << std::endl;
  
  fuse_base::help_options ();
}



int cvsfs_fuse::change_view (const char *viewspec, size_t size)
{
  mount_view.purge_layers ();

  projects.clear ();

  size_t pos = 0;
  
  while (pos < size)
  {
    const char *ptr = &viewspec[pos];
  
    if (*ptr != '\0')
    {
      int result = add_repository (ptr);
      if (result < 0)
        return result;
    }
    pos += strlen (ptr) + 1;
  }

  return 0;
}



int cvsfs_fuse::get_view (char *viewspec, size_t maxsize) const
{
  project_list::const_iterator iter;
  size_t size = 0;
  char *ptr = viewspec;

  for (iter = projects.begin (); iter != projects.end (); ++iter)
  {
    size += (*iter).length () + 1;

    // maxsize == 0 --> only obtain needed size
    if (maxsize != 0)
    {
      if (size > maxsize)
        return -ERANGE;

      strcpy (ptr, (*iter).c_str ());
    
      ptr = &viewspec[size];
    }
  }

  return size;
}



// This is a big function parsing one repository line
// The layout is:
//   [:<method>:][<userid>[:<passord>]@]<server>[:port]<cvsroot> [<repository> [<label>]]
//
// Samples:
//   :pserver:anonymous@127.0.0.1:2401/var/cvsroot . HEAD
//   :pserver:humba:xyz@127.0.0.1/var/cvsroot project HEAD
//   :pserver:127.0.0.1/var/cvsroot
//
int cvsfs_fuse::add_repository (const std::string & spec)
{
  std::string argument = spec;
  std::string::size_type pos;
  std::string method;
  std::string userid;
  std::string password;
  std::string server;
  std::string port;
  std::string cvsroot;
  std::string repository;
  std::string label;
  bool password_obtained = false;

  system_log *log = system_log::instance ();

  // prefixed by access method ?
  if (argument[0] == ':')
  {
    // access method given
    argument.erase (0, 1);
    pos = argument.find (':');
    if (pos == std::string::npos)
    {
      log->error << "add_repository: Connection method not properly terminated." << std::endl;

      return -EFAULT;
    }

    method = argument;
    method.erase (pos);
    argument.erase (0, pos + 1);
  }
  else
    method = "pserver";

  pos = argument.find ('@');
  if (pos != std::string::npos)
  {
    // at least a user ID is given
    userid = argument;
    userid.erase (pos);
    argument.erase (0, pos + 1);
    
    pos = userid.find (':');
    if (pos != std::string::npos)
    {
      // password given
      password = userid;
      password.erase (0, pos + 1);
      userid.erase (pos);
      password_obtained = true;
    }
  }
  else
    userid = "anonymous";
  
  pos = argument.find ('/');
  if (pos == std::string::npos)
  {
    log->error << "add_repository: CVS root not given." << std::endl;

    return -EFAULT;
  }

  server = argument;
  server.erase (pos);
  argument.erase (0, pos);	// keep slash !
  
  pos = server.find (':');
  if (pos != std::string::npos)
  {
    // port number given
    port = server;
    port.erase (0, pos + 1);
    server.erase (pos);
  }

  cvsroot = argument;
  pos = argument.find (' ');
  if (pos != std::string::npos)
  {
    cvsroot.erase (pos);
    argument.erase (0, pos + 1);

    while (isblank (*(argument.begin ())))
      argument.erase (0, 1);

    repository = argument;
    pos = argument.find (' ');
    if (pos != std::string::npos)
    {
      repository.erase (pos);
      argument.erase (0, pos + 1);

      while (isblank (*(argument.begin ())))
        argument.erase (0, 1);

      label = argument;
    }
    else
      label = "HEAD";
  }
  else
  {
    repository = ".";
    label = "HEAD";
  }

  if ((!password_obtained) && (method == "pserver"))
  {
    // here we may try to obtain the password from ~/.cvspass ???
  }

  // ok - now we have all the data we need - lets allocate the repository view
  log->debug << "CVS view mount request:" << std::endl
             << "  Method ..... " << method << std::endl
	     << "  User ID .... " << userid << std::endl
	     << "  Password ... " << password << std::endl
	     << "  Server ..... " << server << std::endl
	     << "  Port ....... " << port << std::endl
	     << "  CVS root ... " << cvsroot << std::endl
	     << "  Project .... " << repository << std::endl
	     << "  Label ...... " << label << std::endl;

  cvs_connection_factory *factory = cvs_connection_factory::instance ();

  if (method == "pserver")
  {
    cvs_connection *conn = factory->get_connection (cvs_connection_factory::PSERVER,
						    server, port, cvsroot, repository,
						    userid, password);

    if (!conn)
    {
      log->error << "add_repository: Unable to establish connection to '" << server << "' - skipped." << std::endl;
    
      return -EFAULT;
    }

    std::string cache_location = cache_base + "/" + server + cvsroot;

    // allocate the repository
    cvs_repository *repo = new cvs_repository (cache_location, conn, repository, label);

    // load tree info if available
    repo->startup ();

    // add the filesystem layers to the view
    mount_view.append_layer (repo->create_checkout_layer ());
    mount_view.append_layer (repo->create_cvs_layer ());

    // add to list - for info
    projects.push_back (spec);

    return 0;
  }

  log->error << "add_repository: Unknown connection method '" << method << "' - skipped." << std::endl;

  return -EFAULT;	// unsupported method
}



int cvsfs_fuse::handle_response (const std::string &response, char *buffer, size_t maxsize)
{
  if (maxsize != 0)
  {
    if (maxsize < response.length ())
      return -ERANGE;

    strcpy (buffer, response.c_str ());
  }

  return response.length () + 1;
}



int cvsfs_fuse::getattr (const char *path, struct stat *stbuf)
{
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::getattr: path = " << path << std::endl;

  memset (stbuf, 0, sizeof (struct stat));
  
  if (strcmp (path, "/") == 0)
  {
    // root directory
    stbuf->st_mode = S_IFDIR | 0755;
    stbuf->st_nlink = 2;
    
    return 0;
  }

  if (mount_view.getattr (path, stbuf) >= 0)
  {
    // override uid/gid values
    stbuf->st_uid = uid;
    stbuf->st_gid = gid;

    return 0;
  }
  
  return -ENOENT;
}



int cvsfs_fuse::readlink (const char *path, char *buffer, size_t size)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::readlink: path = " << path << std::endl;

  if ((result = mount_view.readlink (path, buffer, size)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::getdir (const char *path, fuse_dirh_t h, fuse_dirfil_t filler)
{
  directory::iterator iter;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::getdir: path = " << path << std::endl;
  
  directory *dir = mount_view.readdir (path);
  
  for (iter = dir->begin (); iter != dir->end (); ++iter)
  {
    int result;
  
    if ((result = filler (h, (*iter).c_str (), 0)) != 0)
      return result;
  }

  return 0;
}



int cvsfs_fuse::mknod (const char *path, mode_t mode, dev_t rdev)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::mknod: path = " << path << ", mode = " << mode << std::endl;

  if ((result = mount_view.mknod (path, mode, rdev)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::mkdir (const char *path, mode_t mode)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::mkdir: path = " << path << ", mode = " << mode << std::endl;

  if ((result = mount_view.mkdir (path, mode)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::unlink (const char *path)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::unlnk: path = " << path << std::endl;

  if ((result = mount_view.unlink (path)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::rmdir (const char *path)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::rmdir: path = " << path << std::endl;

  if ((result = mount_view.rmdir (path)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::symlink (const char *path, const char *dest)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::symlink: path = " << path << ", destination = " << dest << std::endl;

  if ((result = mount_view.symlink (path, dest)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::rename (const char *path, const char *newpath)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::rename: path = " << path << ", newpath = " << newpath << std::endl;

  if ((result = mount_view.rename (path, newpath)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::link (const char *path, const char *dest)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::link: path = " << path << ", destination = " << dest << std::endl;

  if ((result = mount_view.link (path, dest)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::chmod (const char *path, mode_t mode)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::chmod: path = " << path << ", mode = " << mode << std::endl;

  if ((result = mount_view.chmod (path, mode)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::chown (const char *path, uid_t uid, gid_t gid)
{
  // cvsfs does not allow uid/gid setting due cvs does not support it
  return -EPERM;
}



int cvsfs_fuse::truncate (const char *path, off_t limit)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::truncate: path = " << path << ", to-size = " << limit << std::endl;

  if ((result = mount_view.truncate (path, limit)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::utime (const char *path, struct utimbuf *buffer)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::utime: path = " << path << std::endl;

  if ((result = mount_view.utime (path, buffer)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::open (const char *path, int flags)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::open: path = " << path << ", flags = " << flags << std::endl;

  if ((result = mount_view.open (path, flags)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::read (const char *path, char *buffer, size_t size, off_t offset)
{
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::read: path = " << path << std::endl;

  return mount_view.read (path, buffer, size, offset);
}



int cvsfs_fuse::write (const char *path, const char *buffer, size_t size, off_t offset)
{
  system_log *log = system_log::instance ();

  log->debug << "cvsfs_fuse::write: path = " << path << std::endl;

  return mount_view.write (path, buffer, size, offset);
}



int cvsfs_fuse::statfs (const char *path, struct statfs *stbuf)
{
  int result;

  if ((result = mount_view.statfs (path, stbuf)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::flush (const char *path)
{
  int result;

  if ((result = mount_view.flush (path)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::release (const char *path, int flags)
{
  int result;

  if ((result = mount_view.release (path, flags)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::fsync (const char *path, int isdatasync)
{
  int result;

  if ((result = mount_view.fsync (path, isdatasync)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::setxattr (const char *path, const char *name, const char *value, size_t size, int flags)
{
  int result;
  system_log *log = system_log::instance ();

  log->debug << "setxattr: path=" << path << ", name=" << name << ", size=" << size << ", flags=" << flags << std::endl;

  if ((strcmp (path, "/") == 0) && (strcmp (name, "view") == 0))
    return change_view (value, size);

  if ((result = mount_view.setxattr (path, name, value, size, flags)) < 0)
    return result;

  return 0;
}



int cvsfs_fuse::getxattr (const char *path, const char *name, char *value, size_t size)
{
  system_log *log = system_log::instance ();

  log->debug << "getxattr: path=" << path << ", name=" << name << ", size=" << size << std::endl;

  if ((strcmp (path, "/") == 0) && (strcmp (name, "view") == 0))
    return get_view (value, size);

  return mount_view.getxattr (path, name, value, size);
}



int cvsfs_fuse::listxattr (const char *path, char *list, size_t size)
{
  return mount_view.listxattr (path, list, size);
}



int cvsfs_fuse::removexattr (const char *path, const char *name)
{
  int result;

  if ((result = mount_view.removexattr (path, name)) < 0)
    return result;

  return 0;
}
