/***************************************************************************
 *   Copyright (C) 2004 by Petric Frank                                    *
 *   pfrank@gmx.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 *   This layer expects all file entries have appended its version         *
 *   number, i.e filename@@1.5                                             *
 *   The example consists of two parts - a file name (= filename) and      *
 *   the version part (1.5). The version part is the version from where    *
 *   it was checked out from the repository.                               *
 *                                                                         *
 *   The files stored in the file system are copies from the files         *
 *   located in the repository.                                            *
 *   In the view from the kernel side this layer is read only. Stores are  *
 *   only done by access of the files which is done transparently.         *
 ***************************************************************************/

#include "fs-layer-cvs.h"
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <utime.h>
#include <fcntl.h>
#include <unistd.h>
#include "cvs-repository.h"



static char *LayerAttrName	= "layer";
static char *LayerAttrValue	= "CVS";



fs_layer_cvs::fs_layer_cvs (const std::string & base, cvs_repository *rep)
: fs_layer (base), repos (rep)
{
}



fs_layer_cvs::~fs_layer_cvs ()
{
  repos->decrement_refcount ();
}



directory *fs_layer_cvs::readdir (const std::string &path) const
{
  // reading is done by accessing the CVS repository
  return repos->readdir (path);
}



int fs_layer_cvs::getattr (const std::string &path, struct stat *stbuf) const
{
  // The things become more complicated. Due the name given does not contain
  // the version requested i have to ask a local directory of the CVS repository.
  int result = repos->getattr (path, stbuf);

  // CVS layer is always read only !
  if ((result >= 0) && S_ISREG (stbuf->st_mode))
    stbuf->st_mode &= ~(S_IWUSR | S_IWGRP | S_IWOTH);

  return result;
}



int fs_layer_cvs::readlink (const std::string &path, char *buffer, size_t size) const
{
  // (symblic) links are not supported in this filesystem
  return -EINVAL;
}



int fs_layer_cvs::mknod (const std::string &path, mode_t mode, dev_t rdev)
{
  // allocation via filesystem is not allowed
  // must be done by VFS during check-out
  return -EINVAL;
}



int fs_layer_cvs::mkdir (const std::string &path, mode_t mode)
{
  // creation of subdirectories via filesystem is not allowed
  // must be done by VFS during check-out
  return -EROFS;
}



int fs_layer_cvs::unlink (const std::string &path)
{
  // this is a readonly filesystem - deleting enties is denied
  return -EROFS;
}



int fs_layer_cvs::rmdir (const std::string &path)
{
  // this is a readonly filesystem - deleting enties is denied
  return -EROFS;
}



int fs_layer_cvs::symlink (const std::string &path, const std::string &dest)
{
  // (symblic) links are not supported in this filesystem
  return -EINVAL;
}



int fs_layer_cvs::rename (const std::string &path, const std::string &newpath)
{
  // rename is not supported by CVS
  return -EINVAL;
}



int fs_layer_cvs::link (const std::string &path, const std::string &dest)
{
  // (symblic) links are not supported by CVS
  return -EINVAL;
}



int fs_layer_cvs::chmod (const std::string &path, mode_t mode)
{
  // this is a readonly filesystem - deleting enties is denied
  return -EROFS;
}



int fs_layer_cvs::chown (const std::string &path, uid_t uid, gid_t gid)
{
  // ownership changes are not supported by CVS
  return -EINVAL;
}



int fs_layer_cvs::truncate (const std::string &path, off_t limit)
{
  // this is a readonly filesystem - changing enties is denied
  return -EROFS;
}



int fs_layer_cvs::utime (const std::string &path, struct utimbuf *buffer)
{
  // this is a readonly filesystem - modifications are denied
  return -EROFS;
}



int fs_layer_cvs::open (const std::string &path, int flags)
{
  // open task: check flags because this is a r/o filesystem
  std::string realpath;
  struct stat stbuf;
  int result;

  if ((result = getattr (path, &stbuf)) < 0)
    return result;

  if (path.find ("@@") != std::string::npos)
    return fs_layer::open (realpath, flags);

  if (!repos->get_fullname (path, &realpath))
    return -ENOENT;

  return fs_layer::open (realpath, flags);
}



int fs_layer_cvs::read (const std::string &path, char *buffer, size_t size, off_t offset) const
{
  std::string realpath;
  struct stat stbuf;
  int result;
  
  if ((result = getattr (path, &stbuf)) < 0)
    return result;

  if (path.find ("@@") != std::string::npos)
    return fs_layer::read (path, buffer, size, offset);

  if (!repos->get_fullname (path, &realpath))
    return -ENOENT;

  return fs_layer::read (realpath, buffer, size, offset);
}



int fs_layer_cvs::write (const std::string &path, const char *buffer, size_t size, off_t offset)
{
  // this is a readonly filesystem - writing is denied
  return -EROFS;
}



int fs_layer_cvs::statfs (const std::string &path, struct statfs *stbuf) const
{
  std::string realpath;
  
  if (!repos->get_fullname (path, &realpath))
    return -ENOENT;

  return fs_layer::statfs (realpath, stbuf);
}



int fs_layer_cvs::setxattr (const std::string &path, const std::string &name, const std::string &value, size_t size, int flags)
{
  // this is a readonly filesystem - modifications are denied
  return -EROFS;
}



int fs_layer_cvs::getxattr (const std::string &path, const std::string &name, char *value, size_t size) const
{
  if (name == LayerAttrName)
  {
    if (size == 0)
      return strlen (LayerAttrValue) + 1;

    if (size <= strlen (LayerAttrValue))
      return -ERANGE;
      
    strcpy (value, LayerAttrValue);
    
    return strlen (LayerAttrValue) + 1;
  }
  
  if (name == "checkout")
    return repos->checkout_file (path, value, size);

  if (name == "getlog")
    return repos->get_log (path, value, size);

  std::string realpath;
  
  if (!repos->get_fullname (path, &realpath))
    return -EOPNOTSUPP;

  if (name == "version")
  {
    std::string::size_type pos = realpath.find ("@@");
    if (pos != std::string::npos)
      realpath.erase (0, pos + 2);
    else
      realpath.clear ();

    if (size == 0)
      return realpath.length () + 1;

    if (size <= realpath.length ())
      return -ERANGE;
	
    strcpy (value, realpath.c_str ());
    
    return realpath.length () + 1;
  }

  return fs_layer::getxattr (realpath, name.c_str (), value, size);
}



int fs_layer_cvs::listxattr (const std::string &path, char *list, size_t size) const
{
  std::string realpath;
  
  if (!repos->get_fullname (path, &realpath))
    return -ENOENT;

  return fs_layer::listxattr (realpath, list, size);
}



int fs_layer_cvs::removexattr (const std::string &path, const std::string &name)
{
  // this is a readonly filesystem - modifications are denied
  return -EROFS;
}
